@@ -17,7 +17,7 @@ void mail_index_sync_init_expunge_handlers(struct mail_index_sync_map_ctx *ctx)
 	void **contexts;
 	struct mail_index_expunge_handler eh;
 	unsigned int ext_count, id_map_count;
-	unsigned int rext_count, context_count;
+	unsigned int rext_count, context_count, count;
 	uint32_t idx_ext_id, map_ext_id;
 
 	if (!array_is_created(&ctx->view->map->extensions))
@@ -35,17 +35,17 @@ void mail_index_sync_init_expunge_handlers(struct mail_index_sync_map_ctx *ctx)
 	contexts = array_get_modifiable(&ctx->extra_contexts, &context_count);
 
 	i_assert(context_count >= rext_count);
-	for (idx_ext_id = 0; idx_ext_id < rext_count; idx_ext_id++) {
-		map_ext_id = idx_ext_id >= id_map_count ? (uint32_t)-1 :
-			id_map[idx_ext_id];
-		if (rext[idx_ext_id].expunge_handler == NULL ||
-		    (map_ext_id == (uint32_t)-1))
+	count = I_MIN(rext_count, id_map_count);
+	for (idx_ext_id = 0; idx_ext_id < count; idx_ext_id++) {
+		if (rext[idx_ext_id].expunge_handler == NULL)
+			continue;
+		map_ext_id = id_map[idx_ext_id];
+		if (map_ext_id == (uint32_t)-1)
 			continue;
 
 		eh.handler = rext[idx_ext_id].expunge_handler;
 		eh.sync_context = &contexts[idx_ext_id];
-		eh.record_offset = map_ext_id == (uint32_t)-1 ? 0 :
-			ext[map_ext_id].record_offset;
+		eh.record_offset = ext[map_ext_id].record_offset;
 		array_push_back(&ctx->expunge_handlers, &eh);
 	}
 	ctx->expunge_handlers_set = TRUE;
@@ -364,6 +364,8 @@ mail_index_sync_ext_init_new(struct mail_index_sync_map_ctx *ctx,
 	buffer_t *hdr_buf;
 	uint32_t ext_map_idx;
 
+	i_assert(mail_index_ext_name_is_valid(name));
+
 	/* be sure to get a unique mapping before we modify the extensions,
 	   otherwise other map users will see the new extension but not the
 	   data records that sync_ext_reorder() adds. */
@@ -406,41 +408,6 @@ mail_index_sync_ext_init_new(struct mail_index_sync_map_ctx *ctx,
 	*ext_map_idx_r = ext_map_idx;
 }
 
-void mail_index_sync_ext_init(struct mail_index_sync_map_ctx *ctx,
-			      const char *name, bool fix_size,
-			      uint32_t *ext_map_idx_r)
-{
-	struct mail_index_map *map = ctx->view->map;
-	const struct mail_index_registered_ext *rext;
-	struct mail_index_ext_header ext_hdr;
-	struct mail_transaction_ext_intro u;
-	uint32_t ext_id;
-
-	if (!mail_index_ext_lookup(ctx->view->index, name, &ext_id))
-		i_unreached();
-	rext = array_idx(&ctx->view->index->extensions, ext_id);
-
-	if (mail_index_map_lookup_ext(map, name, ext_map_idx_r)) {
-		if (!fix_size)
-			return;
-
-		/* make sure it's the expected size */
-		i_zero(&u);
-		u.hdr_size = rext->hdr_size;
-		u.record_size = rext->record_size;
-		u.record_align = rext->record_align;
-		sync_ext_resize(&u, *ext_map_idx_r, ctx, FALSE);
-	} else {
-		i_zero(&ext_hdr);
-		ext_hdr.name_size = strlen(name);
-		ext_hdr.hdr_size = rext->hdr_size;
-		ext_hdr.record_size = rext->record_size;
-		ext_hdr.record_align = rext->record_align;
-		mail_index_sync_ext_init_new(ctx, name, &ext_hdr,
-					     ext_map_idx_r);
-	}
-}
-
 int mail_index_sync_ext_intro(struct mail_index_sync_map_ctx *ctx,
 			      const struct mail_transaction_ext_intro *u)
 {
