@@ -205,7 +205,7 @@ void index_mail_parse_header_init(struct index_mail *mail,
 	index_mail_filter_stream_destroy(mail);
 	i_assert(!mail->data.header_parser_initialized);
 
-	mail->header_seq = data->seq;
+	mail->header_seq = mail->mail.mail.seq;
 	if (mail->header_data == NULL) {
 		mail->header_data = buffer_create_dynamic(default_pool, 4096);
 		i_array_init(&mail->header_lines, 32);
@@ -423,7 +423,7 @@ static void index_mail_init_parser(struct index_mail *mail)
 		data->parser_input = NULL;
 		if (message_parser_deinit_from_parts(&data->parser_ctx, &parts, &error) < 0) {
 			index_mail_set_message_parts_corrupted(&mail->mail.mail, error);
-			data->parts = NULL;
+			index_mail_parts_reset(mail);
 		}
 		if (data->parts == NULL || data->parts != parts) {
 			/* The previous parsing didn't finish, so we're
@@ -681,7 +681,7 @@ index_mail_get_raw_headers(struct index_mail *mail, const char *field,
 			mail_set_aborted(&mail->mail.mail);
 			return -1;
 		}
-		if (mail->header_seq != mail->data.seq ||
+		if (mail->header_seq != mail->mail.mail.seq ||
 		    index_mail_header_is_parsed(mail, field_idx) < 0) {
 			/* parse */
 			const char *reason = index_mail_cache_reason(_mail,
@@ -935,6 +935,7 @@ int index_mail_get_header_stream(struct mail *_mail,
 		/* we have to parse the header. */
 		const char *reason =
 			index_mail_cache_reason(_mail, "bodystructure");
+		mail->data.access_reason_code = "mail:header_fields";
 		if (index_mail_parse_headers(mail, headers, reason) < 0)
 			return -1;
 	}
@@ -972,6 +973,7 @@ int index_mail_get_header_stream(struct mail *_mail,
 			"%u/%u headers not cached (first=%s)",
 			not_found_count, headers->count, headers->name[first_not_found]));
 	}
+	mail->data.access_reason_code = "mail:header_fields";
 	if (mail_get_hdr_stream_because(_mail, NULL, reason, &input) < 0)
 		return -1;
 
