@@ -491,13 +491,15 @@ mail_duplicate_transaction_free(struct mail_duplicate_transaction **_trans)
 	i_assert(trans->db->transaction_count > 0);
 	trans->db->transaction_count--;
 
-	iter = hash_table_iterate_init(trans->hash);
-	while (hash_table_iterate(iter, trans->hash, &d, &d))
-		mail_duplicate_unlock(trans, d);
-	hash_table_iterate_deinit(&iter);
+	if (hash_table_is_created(trans->hash)) {
+		iter = hash_table_iterate_init(trans->hash);
+		while (hash_table_iterate(iter, trans->hash, &d, &d))
+			mail_duplicate_unlock(trans, d);
+		hash_table_iterate_deinit(&iter);
+		hash_table_destroy(&trans->hash);
+	}
 	i_assert(trans->id_lock_count == 0);
 
-	hash_table_destroy(&trans->hash);
 	event_unref(&trans->event);
 	pool_unref(&trans->pool);
 }
@@ -722,14 +724,17 @@ mail_duplicate_db_init(struct mail_user *user, const char *name)
 
 	e_debug(db->event, "Initialize");
 
+	db->user = user;
+
 	if (mail_user_get_home(user, &home) <= 0) {
 		e_error(db->event, "User %s doesn't have home dir set, "
 			"disabling duplicate database", user->username);
+		return db;
 	}
 
-	db->user = user;
-	db->path = home == NULL ? NULL :
-		i_strconcat(home, "/.dovecot.", name, NULL);
+	i_assert(home != NULL);
+
+	db->path = i_strconcat(home, "/.dovecot.", name, NULL);
 	db->dotlock_set = default_mail_duplicate_dotlock_set;
 
 	lock_dir = mail_user_get_volatile_dir(user);
