@@ -124,21 +124,24 @@ virtual_backend_box_lookup_name(struct virtual_mailbox *mbox, const char *name)
 	return NULL;
 }
 
-struct virtual_backend_box *
-virtual_backend_box_lookup(struct virtual_mailbox *mbox, uint32_t mailbox_id)
+bool virtual_backend_box_lookup(struct virtual_mailbox *mbox, uint32_t mailbox_id,
+				struct virtual_backend_box **bbox_r)
 {
 	struct virtual_backend_box *const *bboxes;
 	unsigned int i, count;
 
+	*bbox_r = NULL;
 	if (mailbox_id == 0)
-		return NULL;
+		return FALSE;
 
 	bboxes = array_get(&mbox->backend_boxes, &count);
 	for (i = 0; i < count; i++) {
-		if (bboxes[i]->mailbox_id == mailbox_id)
-			return bboxes[i];
+		if (bboxes[i]->mailbox_id == mailbox_id) {
+			*bbox_r = bboxes[i];
+			return TRUE;
+		}
 	}
-	return NULL;
+	return FALSE;
 }
 
 static bool virtual_mailbox_is_in_open_stack(struct virtual_storage *storage,
@@ -193,9 +196,6 @@ static int virtual_backend_box_alloc(struct virtual_mailbox *mbox,
 	ns = mail_namespace_find(user->namespaces, mailbox);
 	bbox->box = mailbox_alloc(ns->list, mailbox, flags);
 	MODULE_CONTEXT_SET(bbox->box, virtual_storage_module, bbox);
-	mailbox_set_reason(bbox->box, mbox->box.reason == NULL ?
-		t_strdup_printf("virtual mailbox %s", mailbox_get_vname(&mbox->box)) :
-		t_strdup_printf("virtual mailbox %s: %s", mailbox_get_vname(&mbox->box), mbox->box.reason));
 
 	if (bbox == mbox->save_bbox) {
 		/* Assume that the save_bbox exists, whether or not it truly
@@ -860,11 +860,13 @@ static bool virtual_is_inconsistent(struct mailbox *box)
 static int
 virtual_list_index_has_changed(struct mailbox *box ATTR_UNUSED,
 			       struct mail_index_view *list_view ATTR_UNUSED,
-			       uint32_t seq ATTR_UNUSED, bool quick ATTR_UNUSED)
+			       uint32_t seq ATTR_UNUSED, bool quick ATTR_UNUSED,
+			       const char **reason_r)
 {
 	/* we don't have any quick and easy optimizations for tracking
 	   virtual folders. ideally we'd completely disable mailbox list
 	   indexes for them, but this is the easiest way to do it for now. */
+	*reason_r = "Virtual indexes always change";
 	return 1;
 }
 
@@ -930,6 +932,7 @@ struct mailbox virtual_mailbox = {
 		virtual_search_deinit,
 		virtual_search_next_nonblock,
 		virtual_search_next_update_seq,
+		index_storage_search_next_match_mail,
 		virtual_save_alloc,
 		virtual_save_begin,
 		virtual_save_continue,
