@@ -27,6 +27,7 @@ static void
 stats_metric_event(struct metric *metric, struct event *event, pool_t pool);
 static struct metric *
 stats_metric_sub_metric_alloc(struct metric *metric, const char *name, pool_t pool);
+static void stats_metric_free(struct metric *metric);
 
 static void stats_exporters_add_set(struct stats_metrics *metrics,
 				    const struct stats_exporter_settings *set)
@@ -90,8 +91,8 @@ stats_metric_alloc(pool_t pool, const char *name,
 	metric->duration_stats = stats_dist_init();
 	metric->fields_count = str_array_length(fields);
 	if (metric->fields_count > 0) {
-	    metric->fields = p_new(pool, struct metric_field,
-				   metric->fields_count);
+		metric->fields = p_new(pool, struct metric_field,
+				       metric->fields_count);
 		for (unsigned int i = 0; i < metric->fields_count; i++) {
 			metric->fields[i].field_key = p_strdup(pool, fields[i]);
 			metric->fields[i].stats = stats_dist_init();
@@ -187,6 +188,29 @@ stats_metrics_find(struct stats_metrics *metrics,
 	return NULL;
 }
 
+static bool
+stats_metrics_check_for_exporter(struct stats_metrics *metrics, const char *name)
+{
+	struct exporter *exporter;
+
+	/* Allow registering metrics with empty/missing exporters. */
+	if (name[0] == '\0')
+		return TRUE;
+
+	if (!array_is_created(&metrics->exporters))
+		return FALSE;
+
+	bool is_found = FALSE;
+	array_foreach_elem(&metrics->exporters, exporter) {
+		if (strcmp(exporter->name, name) == 0) {
+			is_found = TRUE;
+			break;
+		}
+	}
+
+	return is_found;
+}
+
 bool stats_metrics_add_dynamic(struct stats_metrics *metrics,
 			       struct stats_metric_settings *set,
 			       const char **error_r)
@@ -199,8 +223,15 @@ bool stats_metrics_add_dynamic(struct stats_metrics *metrics,
 
 	struct stats_metric_settings *_set =
 		stats_metric_settings_dup(metrics->pool, set);
-        if (!stats_metric_setting_parser_info.check_func(_set, metrics->pool, error_r))
+	if (!stats_metric_setting_parser_info.check_func(_set, metrics->pool, error_r))
 		return FALSE;
+
+	if (!stats_metrics_check_for_exporter(metrics, set->exporter)) {
+		*error_r = t_strdup_printf("Exporter '%s' does not exist.",
+					   set->exporter);
+		return FALSE;
+	}
+
 	stats_metrics_add_set(metrics, _set);
 	return TRUE;
 }
@@ -209,12 +240,14 @@ bool stats_metrics_remove_dynamic(struct stats_metrics *metrics,
 				  const char *name)
 {
 	unsigned int m_idx;
+	bool ret = FALSE;
 	struct metric *m = stats_metrics_find(metrics, name, &m_idx);
 	if (m != NULL) {
 		array_delete(&metrics->metrics, m_idx, 1);
-		return event_filter_remove_queries_with_context(metrics->filter, m);
+		ret = event_filter_remove_queries_with_context(metrics->filter, m);
+		stats_metric_free(m);
 	}
-	return FALSE;
+	return ret;
 }
 
 static void
@@ -320,8 +353,8 @@ stats_metrics_get_event_filter(struct stats_metrics *metrics)
 }
 
 static struct metric *
-stats_metric_get_sub_metric(struct metric *metric,
-			    const struct metric_value *value)
+stats_metric_find_sub_metric(struct metric *metric,
+			     const struct metric_value *value)
 {
 	struct metric *sub_metrics;
 
@@ -364,22 +397,24 @@ stats_metric_sub_metric_alloc(struct metric *metric, const char *name, pool_t po
 
 static bool
 stats_metric_group_by_discrete(const struct event_field *field,
-			       struct metric_value *value)
+			       struct metric_value *value_r)
 {
 	switch (field->value_type) {
 	case EVENT_FIELD_VALUE_TYPE_STR:
-		value->type = METRIC_VALUE_TYPE_STR;
+		value_r->type = METRIC_VALUE_TYPE_STR;
 		/* use sha1 of value to avoid excessive memory usage in case the
 		   actual value is quite long */
 		sha1_get_digest(field->value.str, strlen(field->value.str),
-				value->hash);
+				value_r->hash);
 		return TRUE;
 	case EVENT_FIELD_VALUE_TYPE_INTMAX:
-		value->type = METRIC_VALUE_TYPE_INT;
-		value->intmax = field->value.intmax;
+		value_r->type = METRIC_VALUE_TYPE_INT;
+		value_r->intmax = field->value.intmax;
 		return TRUE;
 	case EVENT_FIELD_VALUE_TYPE_TIMEVAL:
 		return FALSE;
+	case EVENT_FIELD_VALUE_TYPE_STRLIST:
+		return FALSE;
 	}
 
 	i_unreached();
@@ -388,25 +423,26 @@ stats_metric_group_by_discrete(const struct event_field *field,
 /* convert the value to a bucket index */
 static bool
 stats_metric_group_by_quantized(const struct event_field *field,
-				struct metric_value *value,
+				struct metric_value *value_r,
 				const struct stats_metric_settings_group_by *group_by)
 {
 	switch (field->value_type) {
 	case EVENT_FIELD_VALUE_TYPE_STR:
 	case EVENT_FIELD_VALUE_TYPE_TIMEVAL:
+	case EVENT_FIELD_VALUE_TYPE_STRLIST:
 		return FALSE;
 	case EVENT_FIELD_VALUE_TYPE_INTMAX:
 		break;
 	}
 
-	value->type = METRIC_VALUE_TYPE_BUCKET_INDEX;
+	value_r->type = METRIC_VALUE_TYPE_BUCKET_INDEX;
 
 	for (unsigned int i = 0; i < group_by->num_ranges; i++) {
 		if ((field->value.intmax <= group_by->ranges[i].min) ||
 		    (field->value.intmax > group_by->ranges[i].max))
 			continue;
 
-		value->intmax = i;
+		value_r->intmax = i;
 		return TRUE;
 	}
 
@@ -427,6 +463,7 @@ stats_metric_group_by_quantized_label(const struct event_field *field,
 	switch (field->value_type) {
 	case EVENT_FIELD_VALUE_TYPE_STR:
 	case EVENT_FIELD_VALUE_TYPE_TIMEVAL:
+	case EVENT_FIELD_VALUE_TYPE_STRLIST:
 		i_unreached();
 	case EVENT_FIELD_VALUE_TYPE_INTMAX:
 		break;
@@ -446,15 +483,15 @@ stats_metric_group_by_quantized_label(const struct event_field *field,
 static bool
 stats_metric_group_by_get_value(const struct event_field *field,
 				const struct stats_metric_settings_group_by *group_by,
-				struct metric_value *value)
+				struct metric_value *value_r)
 {
 	switch (group_by->func) {
 	case STATS_METRIC_GROUPBY_DISCRETE:
-		if (!stats_metric_group_by_discrete(field, value))
+		if (!stats_metric_group_by_discrete(field, value_r))
 			return FALSE;
 		return TRUE;
 	case STATS_METRIC_GROUPBY_QUANTIZED:
-		if (!stats_metric_group_by_quantized(field, value, group_by))
+		if (!stats_metric_group_by_quantized(field, value_r, group_by))
 			return FALSE;
 		return TRUE;
 	}
@@ -478,58 +515,127 @@ stats_metric_group_by_get_label(const struct event_field *field,
 	i_panic("unknown group-by function %d", group_by->func);
 }
 
+static const char *
+stats_metric_group_by_value_label(const struct event_field *field,
+				  const struct stats_metric_settings_group_by *group_by,
+				  const struct metric_value *value)
+{
+	switch (value->type) {
+	case METRIC_VALUE_TYPE_STR:
+		return field->value.str;
+	case METRIC_VALUE_TYPE_INT:
+		return dec2str(field->value.intmax);
+	case METRIC_VALUE_TYPE_BUCKET_INDEX:
+		return stats_metric_group_by_get_label(field, group_by, value);
+	}
+	i_unreached();
+}
+
+static struct metric *
+stats_metric_get_sub_metric(struct metric *metric,
+			    const struct event_field *field,
+			    const struct metric_value *value,
+			    pool_t pool)
+{
+	struct metric *sub_metric;
+
+	sub_metric = stats_metric_find_sub_metric(metric, value);
+	if (sub_metric != NULL)
+		return sub_metric;
+
+	T_BEGIN {
+		const char *value_label =
+			stats_metric_group_by_value_label(field,
+				&metric->group_by[0], value);
+		sub_metric = stats_metric_sub_metric_alloc(metric, value_label,
+							   pool);
+	} T_END;
+	if (metric->group_by_count > 1) {
+		sub_metric->group_by_count = metric->group_by_count - 1;
+		sub_metric->group_by = &metric->group_by[1];
+	}
+	sub_metric->group_value.type = value->type;
+	sub_metric->group_value.intmax = value->intmax;
+	memcpy(sub_metric->group_value.hash, value->hash, SHA1_RESULTLEN);
+	return sub_metric;
+}
+
+static void
+stats_metric_group_by_field(struct metric *metric, struct event *event,
+			    const struct event_field *field, pool_t pool)
+{
+	struct metric *sub_metric;
+	struct metric_value value;
+
+	if (!stats_metric_group_by_get_value(field, &metric->group_by[0], &value))
+		return;
+
+	if (!array_is_created(&metric->sub_metrics))
+		p_array_init(&metric->sub_metrics, pool, 8);
+	sub_metric = stats_metric_get_sub_metric(metric, field, &value, pool);
+
+	/* sub-metrics are recursive, so each sub-metric can have additional
+	   sub-metrics. */
+	stats_metric_event(sub_metric, event, pool);
+}
+
+static void
+stats_event_get_strlist(struct event *event, const char *name,
+			ARRAY_TYPE(const_string) *strings)
+{
+	if (event == NULL)
+		return;
+
+	const struct event_field *field =
+		event_find_field_nonrecursive(event, name);
+	if (field != NULL) {
+		const char *str;
+		array_foreach_elem(&field->value.strlist, str)
+			array_push_back(strings, &str);
+	}
+	stats_event_get_strlist(event_get_parent(event), name, strings);
+}
+
 static void
 stats_metric_group_by(struct metric *metric, struct event *event, pool_t pool)
 {
-	const struct stats_metric_settings_group_by *group_by = &metric->group_by[0];
 	const struct event_field *field =
-		event_find_field_recursive(event, group_by->field);
-	struct metric *sub_metric;
-	struct metric_value value;
+		event_find_field_recursive(event, metric->group_by[0].field);
 
 	/* ignore missing field */
 	if (field == NULL)
 		return;
 
-	if (!stats_metric_group_by_get_value(field, group_by, &value))
-		return;
+	if (field->value_type != EVENT_FIELD_VALUE_TYPE_STRLIST)
+		stats_metric_group_by_field(metric, event, field, pool);
+	else {
+		/* Handle each string in strlist separately. The strlist needs
+		   to be combined from the event and its parents, as well as
+		   the global event and its parents. */
+		ARRAY_TYPE(const_string) strings;
 
-	if (!array_is_created(&metric->sub_metrics))
-		p_array_init(&metric->sub_metrics, pool, 8);
+		t_array_init(&strings, 8);
+		stats_event_get_strlist(event, metric->group_by[0].field,
+					&strings);
+		stats_event_get_strlist(event_get_global(),
+					metric->group_by[0].field, &strings);
 
-	sub_metric = stats_metric_get_sub_metric(metric, &value);
+		struct event_field str_field = {
+			.value_type = EVENT_FIELD_VALUE_TYPE_STR,
+		};
+		const char *str;
 
-	if (sub_metric == NULL) T_BEGIN {
-		const char *value_label = NULL;
-
-		switch (value.type) {
-		case METRIC_VALUE_TYPE_STR:
-			value_label = field->value.str;
-			break;
-		case METRIC_VALUE_TYPE_INT:
-			value_label = dec2str(field->value.intmax);
-			break;
-		case METRIC_VALUE_TYPE_BUCKET_INDEX:
-			value_label = stats_metric_group_by_get_label(field,
-								      group_by,
-								      &value);
-			break;
+		/* sort strings so duplicates can be easily skipped */
+		array_sort(&strings, i_strcmp_p);
+		array_foreach_elem(&strings, str) {
+			if (str_field.value.str == NULL ||
+			    strcmp(str_field.value.str, str) != 0) {
+				str_field.value.str = str;
+				stats_metric_group_by_field(metric, event,
+							    &str_field, pool);
+			}
 		}
-
-		sub_metric = stats_metric_sub_metric_alloc(metric, value_label,
-							   pool);
-		if (metric->group_by_count > 1) {
-			sub_metric->group_by_count = metric->group_by_count - 1;
-			sub_metric->group_by = &metric->group_by[1];
-		}
-		sub_metric->group_value.type = value.type;
-		sub_metric->group_value.intmax = value.intmax;
-		memcpy(sub_metric->group_value.hash, value.hash, SHA1_RESULTLEN);
-	} T_END;
-
-	/* sub-metrics are recursive, so each sub-metric can have additional
-	   sub-metrics. */
-	stats_metric_event(sub_metric, event, pool);
+	}
 }
 
 static void
@@ -545,6 +651,7 @@ stats_metric_event_field(struct event *event, const char *fieldname,
 
 	switch (field->value_type) {
 	case EVENT_FIELD_VALUE_TYPE_STR:
+	case EVENT_FIELD_VALUE_TYPE_STRLIST:
 		break;
 	case EVENT_FIELD_VALUE_TYPE_INTMAX:
 		num = field->value.intmax;
